<?php

namespace FiloBlu\Flow\Model\Actions\From;

/**
 * Class Addtoconfigurable
 * @package FiloBlu\Flow\Model\Actions\From
 */
class Addtoconfigurable
{
    /**
     * @var string
     */
    CONST ENTITY_TYPE_CODE = 'catalog_product';
    /**
     * @var \Magento\Catalog\Model\ProductFactory
     */
    protected $_product_factory;

    /**
     * @var \Magento\Framework\App\ObjectManager
     */
    protected $_objectManager;

    /**
     * Addtoconfigurable constructor.
     * @param \Magento\Framework\Model\Context $context
     * @param \Magento\Framework\Registry $registry
     * @param \FiloBlu\Flow\Helper\Product $helper_product
     * @param \Magento\Catalog\Model\ProductFactory $productFactory
     * @param array $data
     */
    public function __construct(
        \Magento\Framework\Model\Context $context,
        \Magento\Framework\Registry $registry,
        \FiloBlu\Flow\Helper\Product $helper_product,
        \Magento\Catalog\Model\ProductFactory $productFactory,
        array $data = []
    ) {
        $this->helper_product = $helper_product;
        $this->_product_factory = $productFactory;
        $this->_objectManager = \Magento\Framework\App\ObjectManager::getInstance();
    }

    /**
     * @param $simpleProduct
     * @param $value
     * @param $storeId
     * @return bool
     */
    public function process($simpleProduct, $value, $storeId)
    {

        if (empty($value) || !$simpleProduct->getId()) {
            return true;
        }

        $this->_objectManager->get(\Magento\Store\Model\StoreManagerInterface::class)->setCurrentStore($storeId);

        if ($storeId != 0) {
            return "The function is avalaible only for the ADMIN store.";
        }

//        if (!preg_match("/^[0-9a-z_\-\.]+#([0-9a-z_]+([,][0-9a-z_]+)*)+$/i", $value)) {
//            return "Configurable SKU is not valid, please use only 0-9 a-z - . and _";
//        }

        $splitted = explode('#', $value);
        $configurableSku = trim($splitted[0]);
        $superAttributeIds = explode(",", $splitted[1]);
        if (count($splitted) > 2) {
            $parentField = trim($splitted[2]);
        }
        unset($splitted);

        if ($configurableSku == $simpleProduct->getSku()) {
            return "Configurable SKU cannot be equal to Simple SKU";
        }

        //check if the simple product has a value for all the super attributes...
        $eavAttributeModel = $this->_objectManager->create('\Magento\Eav\Model\ResourceModel\Entity\Attribute');
        $resourceProductModel = $this->_objectManager->create('Magento\Catalog\Model\ResourceModel\Product');
        $superValuesMap = [];
        foreach ($superAttributeIds as $attributeCode) {
            $attributeId = is_numeric($attributeCode) ? $attributeCode : $eavAttributeModel->getIdByCode('catalog_product',
                $attributeCode);
            $value = $resourceProductModel->getAttributeRawValue($simpleProduct->getId(), $attributeId, $storeId);
            if (!$value) {
                $message = sprintf("Simple product #%s has no value for super attribute '%s'. All super attributes must have a value in the simple product.",
                    $simpleProduct->getId(), $attributeId);
                return $message;
            }

            $superValuesMap[$attributeId] = $value;
        }

        /* getting the configurable product ... */
        $productId = $this->_product_factory->create()->getIdBySku($configurableSku);
        $configurableProduct = $this->_product_factory->create()->setStoreId($storeId)->load($productId);

        if (!$configurableProduct->getId()) {
            $configurableProduct = $this->_createConfigurableFromSimple($configurableSku, $superAttributeIds,
                $simpleProduct, $storeId);
            if (is_string($configurableProduct)) {
                return $configurableProduct;
            }
        } else {
            $this->_syncAttributesFromSimpleToConfigurable($simpleProduct->getId(), $configurableProduct,
                $configurableProduct->getStoreIds());
        }

        /* Getting current associated children... */
        $childIds = $this->_objectManager->create(\Magento\ConfigurableProduct\Model\Product\Type\Configurable::class)->getChildrenIds($configurableProduct->getId());
        $childIds = $childIds[0];

        /* Check if this simple product is already associated... */
        if (in_array($simpleProduct->getId(), $childIds)) {
            /* Resetting visibility and status... */
            $attributesData = [
                'status' => \Magento\Catalog\Model\Product\Attribute\Source\Status::STATUS_ENABLED,
                'visibility' => \Magento\Catalog\Model\Product\Visibility::VISIBILITY_NOT_VISIBLE
            ];

            if (!empty($parentField)) {
                $attributesData[$parentField] = $configurableSku;
            }

            /*
             * TODO: Optimize avoid calling Action but using direct query instead
             */
            $this->_objectManager->get('Magento\Catalog\Model\Product\Action')->updateAttributes(
                array($simpleProduct->getId()), $attributesData, $storeId
            );

            return true;
        }

        //check if already exists a product with the same super attributes values associated to the configurable product...
        $sameValuesProduct = $configurableProduct->getTypeInstance()->getProductByAttributes($superValuesMap,
            $configurableProduct);
        if (!is_null($sameValuesProduct)) {
            $message = sprintf("Product #%s with same super attributes values of simple #%s is already associated to the configurable product.",
                $sameValuesProduct->getId(), $simpleProduct->getId());
            return $message;
        }

        //adding simple to configurable...
        $childIds[] = $simpleProduct->getId();
        \Magento\Framework\App\ObjectManager::getInstance()->create('Magento\ConfigurableProduct\Model\ResourceModel\Product\Type\Configurable')
            ->saveProducts($configurableProduct, $childIds);

        /* needed for correct index prices calculations: START */
        $shouldSaveConfigurable = false;
        if ($configurableProduct->getData('has_options') == 0) {
            $configurableProduct->setData('has_options', 1);
            $shouldSaveConfigurable = true;
        }

        if ($configurableProduct->getData('required_options') == 0) {
            $configurableProduct->setData('required_options', 1);
            $shouldSaveConfigurable = true;
        }

        if ($shouldSaveConfigurable) {
            $configurableProduct->setAssociatedProductIds($childIds);
            $configurableProduct->save();
        }
        /* needed for correct index prices calculations: END */

        //hide the simple product visibility...
        $attributesData = array(
            "status" => \Magento\Catalog\Model\Product\Attribute\Source\Status::STATUS_ENABLED,
            "visibility" => \Magento\Catalog\Model\Product\Visibility::VISIBILITY_NOT_VISIBLE
        );

        if (!empty($parentField)) {
            $attributesData[$parentField] = $configurableSku;
        }

        \Magento\Framework\App\ObjectManager::getInstance()->get('Magento\Catalog\Model\Product\Action')->updateAttributes(
            array($simpleProduct->getId()), $attributesData, $storeId
        );

        return true;
    }

    /**
     * @param $configurableSku
     * @param $superAttributeIds
     * @param $simpleProduct
     * @param int $storeId
     * @return \Magento\Catalog\Model\Product
     */
    protected function _createConfigurableFromSimple($configurableSku, $superAttributeIds, $simpleProduct, $storeId = 0)
    {
        $configurableProduct = $this->_product_factory->create();
        $configurableProduct->setTypeId(\Magento\ConfigurableProduct\Model\Product\Type\Configurable::TYPE_CODE);
        $configurableProduct->setStoreId($storeId);

        //setting the super attributes to the configurable product...
        $superAttributes = array();
        $eavAttributeModel = \Magento\Framework\App\ObjectManager::getInstance()->create('\Magento\Eav\Model\ResourceModel\Entity\Attribute');
        foreach ($superAttributeIds as $superAttributeId) {
            $attributeId = is_numeric($superAttributeId) ? $superAttributeId : $eavAttributeModel->getIdByCode('catalog_product',
                $superAttributeId);
            $attribute = \Magento\Framework\App\ObjectManager::getInstance()->create('\Magento\Eav\Model\Entity\Attribute');
            $attribute->load($attributeId);
            if (!$attribute->getId()) {
                return sprintf("Super attribute with id #%d not found.", $superAttributeId);
            }

            $configurableAttribute = \Magento\Framework\App\ObjectManager::getInstance()->create('Magento\ConfigurableProduct\Model\Product\Type\Configurable\Attribute')->setProductAttribute($attribute);
            $superAttributes[] = [
                'id' => $configurableAttribute->getId(),
                'label' => $configurableAttribute->getLabel(),
                'position' => $attribute->getPosition(),
                'values' => $configurableAttribute->getPrices() ? $configurableAttribute->getPrices() : array(),
                'attribute_id' => $attribute->getId(),
                'attribute_code' => $attribute->getAttributeCode(),
                'frontend_label' => $attribute->getFrontend()->getLabel(),
                'use_default' => 1
            ];
        }

        $configurableProduct->setCanSaveConfigurableAttributes(true);
        $configurableProduct->setConfigurableAttributesData($superAttributes);

        //copy all the attributes of specified groups from the simple product...
        $attributesToCopy = \Magento\Framework\App\ObjectManager::getInstance()->create('FiloBlu\Flow\Model\System\ProductAttributes')->getAttributesToCopy()->getColumnValues('attribute_code');
        $excludedAttributes = array_diff($attributesToCopy,
            $this->helper_product->getSimpleToConfigurableSyncAttributes());
        $attributesToCopy = array_diff($attributesToCopy, $excludedAttributes);
        foreach ($attributesToCopy as $attributeCode) {
            $configurableProduct->setData($attributeCode, $simpleProduct->getData($attributeCode));
            if ($simpleProduct->hasData($attributeCode)) {
                $configurableProduct->setData($attributeCode, $simpleProduct->getData($attributeCode));
            }
        }

        //setting new attributes...
        if (!$configurableProduct->getName()) {
            $configurableProduct->setName($simpleProduct->getName());
        }

        if (!$configurableProduct->getDescription()) {
            $configurableProduct->setDescription($simpleProduct->getDescription());
        }

        if (!$configurableProduct->getShortDescription()) {
            $configurableProduct->setShortDescription($simpleProduct->getShortDescription());
        }

        if (!$configurableProduct->getWeight()) {
            $configurableProduct->setWeight(0);
        }

        $configurableProduct->setSku($configurableSku);
        $configurableProduct->setStatus(\Magento\Catalog\Model\Product\Attribute\Source\Status::STATUS_DISABLED);
        $configurableProduct->setVisibility(\Magento\Catalog\Model\Product\Visibility::VISIBILITY_BOTH);
        $configurableProduct->setAttributeSetId($simpleProduct->getAttributeSetId());
        $configurableProduct->setTaxClassId($simpleProduct->getTaxClassId());
        $configurableProduct->setPrice(0);
        $configurableProduct->setPriceView(0);
        $configurableProduct->setCategoryIds($simpleProduct->getCategoryIds());
        $configurableProduct->setWebsiteIds($simpleProduct->getWebsiteIds());
        $configurableProduct->setStockData(array('is_in_stock' => 1));
        /* needed for correct index prices calculations: START */
        $configurableProduct->setData('has_options', 1);
        $configurableProduct->setData('required_options', 1);
        $configurableProduct->setUseDefault(1);
        /* needed for correct index prices calculations: END */
        try {
            $configurableProduct->save();
        } catch (\Exception $e) {
            if ($e->getMessage() == 'URL key for specified store already exists.') {
                $urlKey = $configurableProduct->getUrlKey();
                $newUrlKey = $urlKey . "-" . $configurableProduct->getId();
                $configurableProduct->setUrlKey($newUrlKey);
                $configurableProduct->save();
            }
        }

        //copy localized attributes from the other stores...
        $this->_syncAttributesFromSimpleToConfigurable($simpleProduct->getId(), $configurableProduct,
            $configurableProduct->getStoreIds());
        return $configurableProduct;
    }

    /**
     * @param $simpleId
     * @param $configurableProduct
     * @param $storeIds
     * @return bool
     */
    protected function _syncAttributesFromSimpleToConfigurable($simpleId, $configurableProduct, $storeIds)
    {

        $attributesToCopy = $this->helper_product->getSimpleToConfigurableSyncAttributes();

        if (empty($attributesToCopy)) {
            return false;
        }

        $superAttributesCode = array();
        // se il configurabile perde i prodotti semplici associati, torna ad essere un semplice, e a riga 275 darà errore - va gestito il caso
        if ($configurableProduct->getTypeId() != \Magento\Catalog\Model\Product\Type::DEFAULT_TYPE) {
            foreach ($configurableProduct->getTypeInstance()->getConfigurableAttributesAsArray($configurableProduct) as $superAttribute) {
                $superAttributesCode[] = $superAttribute['attribute_code'];
            }
        }

        $attributesToCopy = array_diff($attributesToCopy, $superAttributesCode);
        if (empty($attributesToCopy)) {
            return false;
        }

        $configurableId = $configurableProduct->getId();

        array_unshift($storeIds, "0");

        foreach ($storeIds as $storeId) {

            $simple = $this->_product_factory->create()->setStoreId($storeId)->load($simpleId);
            $attributesData = array();
            foreach ($attributesToCopy as $attributeCode) {
                // need to copy only the values actually present for the stores, excluding "use default"
                if ($this->helper_product->getUpdateConfigurableUrl() && $attributeCode == 'url_key') {
                    $name = $simple->getData('name');
                    $confUrlkey = $configurableProduct->formatUrlKey($name . '-' . $storeId . $configurableProduct->getId());
                    $attributesData[$attributeCode] = $confUrlkey;
                } else {
                    $attributesData[$attributeCode] = $simple->getData($attributeCode);
                }
            }

            \Magento\Framework\App\ObjectManager::getInstance()->get('Magento\Catalog\Model\Product\Action')->updateAttributes(array($configurableId),
                $attributesData, $storeId);
            unset($simple, $attributesData);
        }

        return true;
    }

    /**
     * @param $productId
     */
    protected function _prepareProductForSpecialPrice($productId)
    {
        if (!\Magento\Framework\App\ObjectManager::getInstance()->create('Magento\Catalog\Model\ResourceModel\Product')->getAttributeRawValue($productId,
            'special_price', .0)) {
            \Magento\Framework\App\ObjectManager::getInstance()->get('Magento\Catalog\Model\Product\Action')->updateAttributes(
                array($productId), array('special_price' => null), 0
            );
        }
    }

    /**
     * @param $configurableId
     * @param $storeId
     * @param null $websiteId
     */
    protected function _setConfigurablePrice($configurableId, $storeId, $websiteId = null)
    {

    }

}

